/*
 * Copyright (c) 2008-2009, Maconnect
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code MUST retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Maconnect nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "FVWavExporter.h"
#import <QTKit/QTKit.h>
#import <QuickTime/QuickTime.h>


@implementation FVWavExporter

-(BOOL)exportPath:(NSString*)src toPath:(NSString*)dst duration:(float*)duration
{
	BOOL			ok=NO;
	NSDictionary	*open = [NSDictionary dictionaryWithObjectsAndKeys:
							 src, QTMovieFileNameAttribute,
							 [NSNumber numberWithBool:NO], QTMovieOpenAsyncOKAttribute, nil];
	//opening a movie may be a very long process
	QTMovie			*movie=[[QTMovie alloc] initWithAttributes:open error:nil];
	
	if (!movie)
		return NO;
	
	OSType					subtype=kQTFileTypeWave;
	long					dataFormat=k16BitLittleEndianFormat;
	MovieExportComponent	ci = NULL;
	SoundDescriptionHandle	desc = (SoundDescriptionHandle )NewHandleClear(sizeof(SoundDescription));
	
	ComponentDescription    description;
    description.componentType = 'spit';
    description.componentSubType = subtype;
    description.componentManufacturer = 'soun';
    description.componentFlags = 0;
    description.componentFlagsMask = 0;
    Component				ccc = FindNextComponent(0, &description);
	require(ccc, bail);
	ci = OpenComponent(ccc);
	require(ci, bail);
	
	(**desc).descSize = sizeof(SoundDescription);
	(**desc).sampleSize = 16;
	(**desc).sampleRate = 44100 << 16;
	(**desc).dataFormat = dataFormat;
	(**desc).numChannels = 1;
	require(!MovieExportSetSampleDescription(ci, (SampleDescriptionHandle)desc,SoundMediaType), bail);
	
	Movie					inMovie=[movie quickTimeMovie];
	require(inMovie, bail);
	[@"" writeToFile:dst atomically:NO encoding:NSASCIIStringEncoding error:nil];
	
	AliasHandle				aliasHandle;
	FSRef					destRef;
	Boolean					dir;
	ComponentResult			result;
	
	require(!FSPathMakeRef((UInt8*)[dst fileSystemRepresentation], &destRef, &dir), bail);
	require(!FSNewAliasMinimal(&destRef, &aliasHandle), bail);
	
	*duration=((float)GetMovieDuration(inMovie))/((float)GetMovieTimeScale(inMovie));
	
	result=MovieExportToDataRef(ci, (Handle)aliasHandle, rAliasType, inMovie, nil, 0, 136*GetMovieTimeScale(inMovie));
	require(!result, bail);
	ok=YES;
bail:
	if (movie)
		[movie release];
	if (ci)
		CloseComponent(ci);
	if (desc)
		DisposeHandle((Handle)desc);
	return ok;
}

@end
